@@ -13,6 +13,9 @@
  * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#define _GNU_SOURCE
+#define _FILE_OFFSET_BITS 64
+
 #include "libdevmapper.h"
 #include "log.h"
 
@@ -24,6 +27,9 @@
 #include <errno.h>
 #include <unistd.h>
 #include <libgen.h>
+#include <sys/wait.h>
+#include <unistd.h>
+#include <sys/param.h>
 
 #ifdef HAVE_GETOPTLONG
 #  include <getopt.h>
@@ -38,6 +44,15 @@ extern char *optarg;
 #  define OPTIND_INIT 1
 #endif
 
+#ifndef TEMP_FAILURE_RETRY
+# define TEMP_FAILURE_RETRY(expression) \
+  (__extension__                                                              \
+    ({ long int __result;                                                     \
+       do __result = (long int) (expression);                                 \
+       while (__result == -1L && errno == EINTR);                             \
+       __result; }))
+#endif
+
 #ifdef linux
 #  include "kdev_t.h"
 #else
@@ -47,6 +62,7 @@ extern char *optarg;
 #endif
 
 #define LINE_SIZE 1024
+#define ARGS_MAX 256
 
 #define err(msg, x...) fprintf(stderr, msg "\n", ##x)
 
@@ -56,12 +72,14 @@ extern char *optarg;
 enum {
 	READ_ONLY = 0,
 	COLS_ARG,
+	EXEC_ARG,
 	MAJOR_ARG,
 	MINOR_ARG,
 	NOHEADINGS_ARG,
 	NOOPENCOUNT_ARG,
 	NOTABLE_ARG,
 	OPTIONS_ARG,
+	TARGET_ARG,
 	UUID_ARG,
 	VERBOSE_ARG,
 	VERSION_ARG,
@@ -72,6 +90,8 @@ static int _switches[NUM_SWITCHES];
 static int _values[NUM_SWITCHES];
 static char *_uuid;
 static char *_fields;
+static char *_target;
+static char *_command;
 
 /*
  * Commands
@@ -598,6 +618,102 @@ static int _process_all(int argc, char **argv,
 	return r;
 }
 
+static void _display_dev(struct dm_task *dmt, char *name)
+{
+	struct dm_info info;
+
+	if (dm_task_get_info(dmt, &info))
+		printf("%s\t(%u, %u)\n", name, info.major, info.minor);
+}
+
+static int _mknodes_single(char *name)
+{
+	struct dm_task *dmt;
+	int r = 0;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_MKNODES)))
+		return 0;
+
+	if (!_set_task_device(dmt, name, 1))
+		goto out;
+
+	if (_switches[NOOPENCOUNT_ARG] && !dm_task_no_open_count(dmt))
+		goto out;
+
+	if (!dm_task_run(dmt))
+		goto out;
+
+	r = 1;
+
+      out:
+	dm_task_destroy(dmt);
+	return r;
+}
+
+static int _mknodes(int argc, char **argv, void *data)
+{
+	return _mknodes_single(argc > 1 ? argv[1] : NULL);
+}
+
+static int _exec_command(char *name)
+{
+	int n;
+	static char path[PATH_MAX];
+	static char *args[ARGS_MAX + 1];
+	static int argc = 0;
+	char *c;
+	pid_t pid;
+
+	if (argc < 0)
+		return 0;
+
+	if (!_mknodes_single(name))
+		return 0;
+
+	n = snprintf(path, sizeof(path), "%s/%s", dm_dir(), name);
+	if (n < 0 || n > sizeof(path) - 1)
+		return 0;
+
+	if (!argc) {
+		c = _command;
+		while (argc < ARGS_MAX) {
+			while (*c && isspace(*c))
+				c++;
+			if (!*c)
+				break;
+			args[argc++] = c;
+			while (*c && !isspace(*c))
+				c++;
+			if (*c)
+				*c++ = '\0';
+		}
+
+		if (!argc) {
+			argc = -1;
+			return 0;
+		}
+
+		if (argc == ARGS_MAX) {
+			err("Too many args to --exec\n");
+			argc = -1;
+			return 0;
+		}
+
+		args[argc++] = path;
+		args[argc] = NULL;
+	}
+
+	if (!(pid = fork())) {
+		execvp(args[0], args);
+		exit(127);
+	} else if (pid < (pid_t) 0)
+		return 0;
+
+	TEMP_FAILURE_RETRY(waitpid(pid, NULL, 0));
+
+	return 1;
+}
+
 static int _status(int argc, char **argv, void *data)
 {
 	int r = 0;
@@ -609,6 +725,8 @@ static int _status(int argc, char **argv, void *data)
 	int cmd;
 	struct dm_names *names = (struct dm_names *) data;
 	char *name = NULL;
+	int matched = 0;
+	int ls_only = 0;
 
 	if (data)
 		name = names->name;
@@ -624,6 +742,9 @@ static int _status(int argc, char **argv, void *data)
 	else
 		cmd = DM_DEVICE_STATUS;
 
+	if (!strcmp(argv[0], "ls"))
+		ls_only = 1;
+
 	if (!(dmt = dm_task_create(cmd)))
 		return 0;
 
@@ -636,31 +757,48 @@ static int _status(int argc, char **argv, void *data)
 	if (!dm_task_run(dmt))
 		goto out;
 
-	if (_switches[VERBOSE_ARG])
-		_display_info(dmt);
+	if (!name)
+		name = (char *) dm_task_get_name(dmt);
 
 	/* Fetch targets and print 'em */
 	do {
 		next = dm_get_next_target(dmt, next, &start, &length,
 					  &target_type, &params);
-		if (data && !_switches[VERBOSE_ARG])
-			printf("%s: ", name);
-		if (target_type) {
-			printf("%" PRIu64 " %" PRIu64 " %s %s",
-			       start, length, target_type, params);
+		/* Skip if target type doesn't match */
+		if (_switches[TARGET_ARG] &&
+		    (!target_type || strcmp(target_type, _target)))
+			continue;
+		if (ls_only) {
+			if (!_switches[EXEC_ARG] || !_command ||
+			    _switches[VERBOSE_ARG])
+				_display_dev(dmt, name);
+			next = NULL;
+		} else if (!_switches[EXEC_ARG] || !_command ||
+			   _switches[VERBOSE_ARG]) {
+			if (!matched && _switches[VERBOSE_ARG])
+				_display_info(dmt);
+			if (data && !_switches[VERBOSE_ARG])
+				printf("%s: ", name);
+			if (target_type) {
+				printf("%" PRIu64 " %" PRIu64 " %s %s",
+				       start, length, target_type, params);
+			}
+			printf("\n");
 		}
-		printf("\n");
+		matched = 1;
 	} while (next);
 
-	if (data && _switches[VERBOSE_ARG])
+	if (data && _switches[VERBOSE_ARG] && matched && !ls_only)
 		printf("\n");
 
+	if (matched && _switches[EXEC_ARG] && _command && !_exec_command(name))
+		goto out;
+
 	r = 1;
 
       out:
 	dm_task_destroy(dmt);
 	return r;
-
 }
 
 /* Show target names and their version numbers */
@@ -697,30 +835,6 @@ static int _targets(int argc, char **argv, void *data)
 
 }
 
-static int _mknodes(int argc, char **argv, void *data)
-{
-	struct dm_task *dmt;
-	int r = 0;
-
-	if (!(dmt = dm_task_create(DM_DEVICE_MKNODES)))
-		return 0;
-
-	if (!_set_task_device(dmt, argc > 1 ? argv[1] : NULL, 1))
-		goto out;
-
-	if (_switches[NOOPENCOUNT_ARG] && !dm_task_no_open_count(dmt))
-		goto out;
-
-	if (!dm_task_run(dmt))
-		goto out;
-
-	r = 1;
-
-      out:
-	dm_task_destroy(dmt);
-	return r;
-}
-
 static int _info(int argc, char **argv, void *data)
 {
 	int r = 0;
@@ -835,7 +949,11 @@ static int _display_name(int argc, char **argv, void *data)
 
 static int _ls(int argc, char **argv, void *data)
 {
-	return _process_all(argc, argv, _display_name);
+	if ((_switches[TARGET_ARG] && _target) ||
+	    (_switches[EXEC_ARG] && _command))
+		return _status(argc, argv, data);
+	else
+		return _process_all(argc, argv, _display_name);
 }
 
 /*
@@ -864,11 +982,11 @@ static struct command _commands[] = {
 	{"reload", "<device> [<table_file>]", 0, 2, _load},
 	{"rename", "<device> <new_name>", 1, 2, _rename},
 	{"message", "<device> <sector> <message>", 2, -1, _message},
-	{"ls", "", 0, 0, _ls},
+	{"ls", "[--target <target_type>] [--exec <command>]", 0, 0, _ls},
 	{"info", "[<device>]", 0, 1, _info},
 	{"deps", "[<device>]", 0, 1, _deps},
-	{"status", "[<device>]", 0, 1, _status},
-	{"table", "[<device>]", 0, 1, _status},
+	{"status", "[<device>] [--target <target_type>]", 0, 1, _status},
+	{"table", "[<device>] [--target <target_type>]", 0, 1, _status},
 	{"wait", "<device> [<event_nr>]", 0, 2, _wait},
 	{"mknodes", "[<device>]", 0, 1, _mknodes},
 	{"targets", "", 0, 0, _targets},
@@ -905,22 +1023,24 @@ static struct command *_find_command(const char *name)
 static int _process_switches(int *argc, char ***argv)
 {
 	char *base, *namebase;
-	int ind;
+	static int ind;
 	int c;
 
 #ifdef HAVE_GETOPTLONG
 	static struct option long_options[] = {
-		{"readonly", 0, NULL, READ_ONLY},
-		{"columns", 0, NULL, COLS_ARG},
-		{"major", 1, NULL, MAJOR_ARG},
-		{"minor", 1, NULL, MINOR_ARG},
-		{"noheadings", 0, NULL, NOHEADINGS_ARG},
-		{"noopencount", 0, NULL, NOOPENCOUNT_ARG},
-		{"notable", 0, NULL, NOTABLE_ARG},
-		{"options", 1, NULL, OPTIONS_ARG},
-		{"uuid", 1, NULL, UUID_ARG},
-		{"verbose", 1, NULL, VERBOSE_ARG},
-		{"version", 0, NULL, VERSION_ARG},
+		{"readonly", 0, &ind, READ_ONLY},
+		{"columns", 0, &ind, COLS_ARG},
+		{"exec", 1, &ind, EXEC_ARG},
+		{"major", 1, &ind, MAJOR_ARG},
+		{"minor", 1, &ind, MINOR_ARG},
+		{"noheadings", 0, &ind, NOHEADINGS_ARG},
+		{"noopencount", 0, &ind, NOOPENCOUNT_ARG},
+		{"notable", 0, &ind, NOTABLE_ARG},
+		{"options", 1, &ind, OPTIONS_ARG},
+		{"target", 1, &ind, TARGET_ARG},
+		{"uuid", 1, &ind, UUID_ARG},
+		{"verbose", 1, &ind, VERBOSE_ARG},
+		{"version", 0, &ind, VERSION_ARG},
 		{"", 0, NULL, 0}
 	};
 #else
@@ -969,8 +1089,8 @@ static int _process_switches(int *argc, char ***argv)
 
 	optarg = 0;
 	optind = OPTIND_INIT;
-	while ((c = GETOPTLONG_FN(*argc, *argv, "cCj:m:no:ru:v",
-				  long_options, &ind)) != -1) {
+	while ((ind = -1, c = GETOPTLONG_FN(*argc, *argv, "cCj:m:no:ru:v",
+					    long_options, NULL)) != -1) {
 		if (c == 'c' || c == 'C' || ind == COLS_ARG)
 			_switches[COLS_ARG]++;
 		if (c == 'r' || ind == READ_ONLY)
@@ -995,6 +1115,14 @@ static int _process_switches(int *argc, char ***argv)
 			_switches[UUID_ARG]++;
 			_uuid = optarg;
 		}
+		if ((ind == EXEC_ARG)) {
+			_switches[EXEC_ARG]++;
+			_command = optarg;
+		}
+		if ((ind == TARGET_ARG)) {
+			_switches[TARGET_ARG]++;
+			_target = optarg;
+		}
 		if ((ind == NOHEADINGS_ARG))
 			_switches[NOHEADINGS_ARG]++;
 		if ((ind == NOOPENCOUNT_ARG))
@@ -1061,8 +1189,5 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	dm_lib_release();
-	dm_lib_exit();
-
 	return 0;
 }
